// Jabberer.java
//
// CompSci 141 / CSE 141 / Informatics 101 Spring 2013
// Code Example
//
// The Jabberer class encapsulates one "jabberer": an object that says a word
// repeatedly.  Whenever it's told that it is allowed to speak, it says its
// word, then lets other jabberers speak before getting a turn again.
//
// The jabberers will speak one at a time using a principle similar to that
// used on token ring computer networks.  A general problem faced on a network
// is what to do if two computers want to send a message simultaneously.  A
// token ring network solves the problem by passing around a "token".  When a
// computer has the token, it can send a message, but then relinquishes the
// token to another computer.  It can't send another message until it gets
// the token back at a later time.
//
// Instead of passing around a token, we'll implement the same idea by using
// threads, along with Java's wait/notify mechanism.  Every jabberer will be
// given an object whose monitor will be used by all the jabberers.
// Initially, all the jabberers will wait on the monitor.  At some point, the
// main thread will notify the monitor, waking one of the jabberers.  When
// awakened, the jabberer will state its word, sleep for a quarter of a second
// (to keep the output slow enough to read), notify the monitor (waking another
// of the jabberers), then wait on the monitor.

public class Jabberer implements Runnable
{
	private String word;
	private Object monitor;


	// When you construct a Jabberer, you give it two things: the word
	// you'd like it to say and the object whose monitor will be used
	// to notify this Jabberer when it is allowed to speak.  (In order
	// for this example to work, all Jabberers must be given the same
	// monitor.  For fun, try giving them all different monitors and
	// see what happens.  For even more fun, try to guess what happens,
	// then see what the effect is.)
	public Jabberer(String word, Object monitor)
	{
		this.word = word;
		this.monitor = monitor;
	}


	// For fun, you might want to remove the synchronization and wait/notify
	// usage from this method and see what happens.  If you want a really
	// entertaining sight, try removing the call to sleep(), too!
	public void run()
	{
		try
		{
			// Before this thread is allowed to make any progress, it will
			// have to acquire the monitor's lock.  This will prevent two
			// jabberers from ever simultaneously making progress, which
			// will allow us to be sure that all the other jabberers will
			// be waiting before this jabberer ever calls notify().
			//
			// It should be noted again that the call to monitor.wait()
			// will cause the monitor's lock to be released by this thread.
			// When this thread is subsequently notified and wakes up, it
			// won't actually be able to make any progress until it can
			// re-acquire the lock.  So, the way this code is set up, no
			// jabberer threads will make progress unless they have acquired
			// the lock.  This prevents the problem of a jabberer calling
			// notify() before any of the others have had a chance to call
			// wait().

			synchronized (monitor)
			{
				// As in the previous example, if this thread is interrupted,
				// we'll jump down to the catch for InterruptedException below,
				// which will take us out of this loop.

				while (true)
				{
					monitor.wait();

					Console.output(word + " ");

					Thread.sleep(250);

					monitor.notify();
				}
			}
		}
		catch (InterruptedException e)
		{
			// By placing the try/catch outside of everything, whenever this
			// thread is interrupted, we'll immediately end up here, then fall
			// out of the run() method, which will terminate the thread.  It
			// should be noted that an InterruptedException won't actually be
			// thrown until the thread either calls wait() or sleep(), but this
			// is where it spends the vast majority of its time, so that's
			// a reasonable restriction.
		}
	}
}
